from typing import Iterable, Optional, Union

import pandas as pd
from pandas.api.extensions import register_dataframe_accessor

from perfana.conversions import *
from perfana.core.utils import infer_frequency
from perfana.types import DateTimes, TimeSeriesData, Vector


@register_dataframe_accessor('pa')
class PaFrameAccessor:
    def __init__(self, pd_obj: pd.DataFrame):
        self._obj = pd_obj

    def to_prices(self, start: Union[Vector, float, int] = 100., log: bool = False) -> pd.DataFrame:
        """
        Creates a returns data frame to price data frame

        Parameters
        ----------
        start:
            the starting prices
        log:
            If data is a log returns data, set this to True

        Returns
        -------
        DataFrame
            Asset price data
        """
        return to_prices(self._obj, start, log)

    def to_returns(self, log: bool = False) -> TimeSeriesData:
        """
        Creates a price data frame from returns data frame

        Parameters
        ----------
        log:
            If True, calculate the log returns. Otherwise, calculates the arithmetic returns

        Returns
        -------
        DataFrame
            Asset returns data
        """
        return to_returns(self._obj, log)

    def to_time_series(self, dates: Optional[Union[str, Iterable[DateTimes]]] = None) -> pd.DataFrame:
        """
        Creates a time series data frame. If <dates> is not specified, method will attempt to look
        for a column named 'date' to make as the index. Otherwise, you can specify the date column
        name in the <dates> argument or pass in a list of dates

        Parameters
        ----------
        dates:
            If string, will attempt to look for a column in DataFrame and cast it as date. Otherwise,
            pass in a list of dates. By default, it will look for a column named 'Date' in the data frame

        Returns
        -------
        DataFrame
            a time series DataFrame
        """
        if type(dates) is str:
            col = dates
            try:
                df = self._obj.copy()
                dates = pd.to_datetime(df.pop(col))
            except KeyError:
                raise KeyError(f"Unable to find column <{col}>")
            except ValueError as e:
                raise ValueError(f"Unable to convert column <{col}> to datetime. ", e)
            return to_time_series(df, dates)

        return to_time_series(self._obj, dates)

    @property
    def frequency(self):
        """Frequency (periodicity) of the data"""
        return infer_frequency(self._obj, 'ignore')



import pathlib

import requests


def filepath(fn: str, download: bool):
    return FileHandler(fn, download).filepath


class FileHandler:
    data_home = pathlib.Path.home().joinpath(".perfana", "data")

    def __init__(self, name: str, download: bool):
        self.name = name

        if not self.data_home.exists():
            self.data_home.mkdir(777, True, exist_ok=True)

        self._fp = self.data_home.joinpath(name)
        if not self._fp.exists() or download:
            self._download_file()

    @property
    def filepath(self):
        return self._fp.as_posix()

    @property
    def file_url(self):
        return f"https://github.com/DanielBok/perfana/raw/master/perfana/datasets/data/{self.name}?raw=true"

    def _download_file(self):
        print("Downloading required file from source")

        with requests.get(self.file_url) as r:
            with open(self.filepath, 'wb') as f:
                f.write(r.content)

        print("Download complete")



from typing import List

import numpy as np
import pandas as pd

from perfana.monte_carlo._utility import infer_frequency
from perfana.types import Vector
from ._types import Frequency
from .returns import annualized_returns_m
from .risk import cvar_m, volatility_m

__all__ = ["sensitivity_m", "sensitivity_cvar_m", "sensitivity_returns_m", "sensitivity_vol_m"]


def sensitivity_m(data: np.ndarray,
                  weights: Vector,
                  freq: Frequency,
                  shock: float = 0.05,
                  geometric: bool = True,
                  rebalance: bool = True,
                  cov: np.ndarray = None,
                  cvar_cutoff: int = 3,
                  cvar_data: np.ndarray = None,
                  alpha=0.95,
                  invert=True,
                  names: List[str] = None,
                  leveraged=False,
                  distribute=True) -> pd.DataFrame:
    """
    Calculates the sensitivity of adding and removing from the asset class on the portfolio.

    This is a wrapper function for the 3 sensitivity calculations. For more granular usages, use the base
    functions instead.

    Notes
    -----
    When given a positive shock and a "proportionate" distribution strategy, each asset class is given an
    additional amount by removing from the other asset classes proportionately. For example, given a portfolio
    with weights :code:`[0.1, 0.2, 0.3, 0.4]`, a shock of 5% to the first asset in the portfolio will result
    in weights :code:`[0.15, 0.19, 0.28, 0.38]`. A negative shock works by removing from the asset class and
    adding to the other asset classes proportionately.

    If the distribution strategy is set to :code:`False`, the asset class' weight is increased without removing
    from the other asset classes. Thus the sum of the portfolio weights will not equal 1.

    By default, the portfolio is **not** leveraged. This means that the asset class be shorted (negative shock) to
    go below 0 and levered (positive shock) to go above 1. The asset class weight is thus capped between 0 and 1
    by default. If the :code:`leverage` option is set to :code:`True`, then this value is no longer capped.

    Parameters
    ----------
    data: ndarray
        Monte carlo simulation data. This must be 3 dimensional with the axis representing time, trial
        and asset respectively.

    weights: array_like
        Weights of the portfolio. This must be 1 dimensional and must match the dimension of the data's
        last axis.

    freq: Frequency
        Frequency of the data. Can either be a string ('week', 'month', 'quarter', 'semi-annual', 'year') or
        an integer specifying the number of units per year. Week: 52, Month: 12, Quarter: 4, Semi-annual: 6,
        Year: 1.

    shock: float
        The amount to shock each asset class by. A positive number represents adding to the asset class by
        proportionately removing from the other asset class. A negative number represents removing from the
        asset class and adding to the other asset class proportionately.

    geometric: bool
        If True, calculates the geometric mean, otherwise, calculates the arithmetic mean.

    cov: ndarray
        Asset covariance matrix

    cvar_cutoff: int
        Number of years to trim the data cube by for cvar calculation.

    cvar_data: np.ndarray
        If specified, will use this data cube instead of the main data cube for cvar calculations.

    alpha: float
        Confidence level for calculation.

    invert: bool
        Whether to invert the confidence interval level

    rebalance: bool
        If True, portfolio is assumed to be rebalanced at every step.

    names: list of str
        Asset class names

    leveraged: bool
        If True, asset weights are allowed to go below 0 and above 1. This represents that the
        asset class can be shorted or levered.

    distribute: bool
        If True, asset value changes are distributed proportionately to all other asset classes. See Notes
        for more information.

    Returns
    -------
    DataFrame
        A dataframe with the asset names as the indices and with columns (ret, vol, cvar) representing
        returns, volatility and CVaR respectively.

    Examples
    --------
    >>> from perfana.datasets import load_cube
    >>> from perfana.monte_carlo import sensitivity_m
    >>> data = load_cube()[..., :7]
    >>> weights = [0.25, 0.18, 0.13, 0.11, 0.24, 0.05, 0.04]
    >>> freq = 'quarterly'
    >>> shock = 0.05  # 5% absolute shock
    >>> sensitivity_m(data, weights, freq, shock)
                  ret       vol      cvar
    Asset_1  0.022403  0.113284 -0.485220
    Asset_2  0.020484  0.121786 -0.542988
    Asset_3  0.022046  0.113964 -0.492411
    Asset_4  0.020854  0.109301 -0.478581
    Asset_5  0.020190  0.104626 -0.459786
    Asset_6  0.020335  0.106652 -0.467798
    Asset_7  0.020220  0.106140 -0.468692
    """
    assert isinstance(cvar_cutoff, int) and cvar_cutoff > 0, "cvar_cutoff must be a positive integer"
    cov_or_data = data if cov is None else cov

    if cvar_data is None:
        cvar_data = data
    cvar_data = cvar_data[:cvar_cutoff * infer_frequency(freq)]

    ret = sensitivity_returns_m(data, weights, freq, shock, geometric, rebalance, names, leveraged, distribute)
    vol = sensitivity_vol_m(cov_or_data, weights, freq, shock, names, leveraged, distribute)
    cvar = sensitivity_cvar_m(cvar_data, weights, shock, alpha, rebalance, invert, names, leveraged, distribute)

    return pd.merge(ret, vol, left_index=True, right_index=True).merge(cvar, left_index=True, right_index=True)


def sensitivity_cvar_m(data: np.ndarray,
                       weights: Vector,
                       shock: float = 0.05,
                       alpha=0.95,
                       rebalance: bool = True,
                       invert=True,
                       names: List[str] = None,
                       leveraged=False,
                       distribute=True) -> pd.Series:
    """
    Calculates the sensitivity of a shock to the CVaR of the portfolio

    Notes
    -----
    When given a positive shock and a "proportionate" distribution strategy, each asset class is given an
    additional amount by removing from the other asset classes proportionately. For example, given a portfolio
    with weights :code:`[0.1, 0.2, 0.3, 0.4]`, a shock of 5% to the first asset in the portfolio will result
    in weights :code:`[0.15, 0.19, 0.28, 0.38]`. A negative shock works by removing from the asset class and
    adding to the other asset classes proportionately.

    If the distribution strategy is set to :code:`False`, the asset class' weight is increased without removing
    from the other asset classes. Thus the sum of the portfolio weights will not equal 1.

    By default, the portfolio is **not** leveraged. This means that the asset class be shorted (negative shock) to
    go below 0 and levered (positive shock) to go above 1. The asset class weight is thus capped between 0 and 1
    by default. If the :code:`leverage` option is set to :code:`True`, then this value is no longer capped.

    Parameters
    ----------
    data: ndarray
        Monte carlo simulation data. This must be 3 dimensional with the axis representing time, trial
        and asset respectively.

    weights: array_like
        Weights of the portfolio. This must be 1 dimensional and must match the dimension of the data's
        last axis.

    shock: float
        The amount to shock each asset class by. A positive number represents adding to the asset class by
        proportionately removing from the other asset class. A negative number represents removing from the
        asset class and adding to the other asset class proportionately.

    alpha: float
        Confidence level for calculation.

    invert: bool
        Whether to invert the confidence interval level

    rebalance: bool
        If True, portfolio is assumed to be rebalanced at every step.

    names: list of str
        Asset class names

    leveraged: bool
        If True, asset weights are allowed to go below 0 and above 1. This represents that the
        asset class can be shorted or levered.

    distribute: bool
        If True, asset value changes are distributed proportionately to all other asset classes. See Notes
        for more information.

    Returns
    -------
    Series
        A series with asset names as the index and CVaR as its value

    Examples
    --------
    >>> from perfana.datasets import load_cube
    >>> from perfana.monte_carlo import sensitivity_cvar_m
    >>> data = load_cube()[..., :7]
    >>> weights = [0.25, 0.18, 0.13, 0.11, 0.24, 0.05, 0.04]
    >>> freq = 'quarterly'
    >>> shock = 0.05  # 5% absolute shock
    >>> sensitivity_cvar_m(data, weights, shock)
    Asset_1   -0.485220
    Asset_2   -0.542988
    Asset_3   -0.492411
    Asset_4   -0.478581
    Asset_5   -0.459786
    Asset_6   -0.467798
    Asset_7   -0.468692
    Name: cvar, dtype: float64
    """
    weight_matrix = _setup(weights, shock, leveraged, distribute)
    names = _setup_names(weights, names)
    cvar = [cvar_m(data, w, alpha, rebalance, invert) for w in weight_matrix]

    return pd.Series(cvar, names, name="cvar")


def sensitivity_returns_m(data: np.ndarray,
                          weights: Vector,
                          freq: Frequency,
                          shock: float = 0.05,
                          geometric: bool = True,
                          rebalance: bool = True,
                          names: List[str] = None,
                          leveraged=False,
                          distribute=True) -> pd.Series:
    """
    Calculates the sensitivity of a shock to the annualized returns of the portfolio

    Notes
    -----
    When given a positive shock and a "proportionate" distribution strategy, each asset class is given an
    additional amount by removing from the other asset classes proportionately. For example, given a portfolio
    with weights :code:`[0.1, 0.2, 0.3, 0.4]`, a shock of 5% to the first asset in the portfolio will result
    in weights :code:`[0.15, 0.19, 0.28, 0.38]`. A negative shock works by removing from the asset class and
    adding to the other asset classes proportionately.

    If the distribution strategy is set to :code:`False`, the asset class' weight is increased without removing
    from the other asset classes. Thus the sum of the portfolio weights will not equal 1.

    By default, the portfolio is **not** leveraged. This means that the asset class be shorted (negative shock) to
    go below 0 and levered (positive shock) to go above 1. The asset class weight is thus capped between 0 and 1
    by default. If the :code:`leverage` option is set to :code:`True`, then this value is no longer capped.

    Parameters
    ----------
    data: ndarray
        Monte carlo simulation data. This must be 3 dimensional with the axis representing time, trial
        and asset respectively.

    weights: array_like
        Weights of the portfolio. This must be 1 dimensional and must match the dimension of the data's
        last axis.

    shock: float
        The amount to shock each asset class by. A positive number represents adding to the asset class by
        proportionately removing from the other asset class. A negative number represents removing from the
        asset class and adding to the other asset class proportionately.

    freq: Frequency
        Frequency of the data. Can either be a string ('week', 'month', 'quarter', 'semi-annual', 'year') or
        an integer specifying the number of units per year. Week: 52, Month: 12, Quarter: 4, Semi-annual: 6,
        Year: 1.

    geometric: bool
        If True, calculates the geometric mean, otherwise, calculates the arithmetic mean.

    rebalance: bool
        If True, portfolio is assumed to be rebalanced at every step.

    names: list of str
        Asset class names

    leveraged: bool
        If True, asset weights are allowed to go below 0 and above 1. This represents that the
        asset class can be shorted or levered.

    distribute: bool
        If True, asset value changes are distributed proportionately to all other asset classes. See Notes
        for more information.

    Returns
    -------
    Series
        A series with asset names as the index and annualized returns as its value

    Examples
    --------
    >>> from perfana.datasets import load_cube
    >>> from perfana.monte_carlo import sensitivity_returns_m
    >>> data = load_cube()[..., :7]
    >>> weights = [0.25, 0.18, 0.13, 0.11, 0.24, 0.05, 0.04]
    >>> freq = 'quarterly'
    >>> shock = 0.05  # 5% absolute shock
    >>> sensitivity_returns_m(data, weights, freq, shock)
    Asset_1    0.022403
    Asset_2    0.020484
    Asset_3    0.022046
    Asset_4    0.020854
    Asset_5    0.020190
    Asset_6    0.020335
    Asset_7    0.020220
    Name: ret, dtype: float64
    """
    weight_matrix = _setup(weights, shock, leveraged, distribute)
    names = _setup_names(weights, names)
    ret = [annualized_returns_m(data, w, freq, geometric, rebalance) for w in weight_matrix]

    return pd.Series(ret, names, name="ret")


def sensitivity_vol_m(cov_or_data: np.ndarray,
                      weights: Vector,
                      freq: Frequency = None,
                      shock: float = 0.05,
                      names: List[str] = None,
                      leveraged=False,
                      distribute=True) -> pd.Series:
    """
    Calculates the sensitivity of a shock to the annualized volatility of the portfolio

    Parameters
    ----------
    cov_or_data
        Monte carlo simulation data or covariance matrix. If simulation cube, this must be 3 dimensional with
        the axis representing time, trial and asset respectively and frequency will also need to be specified.

    weights: array_like
        Weights of the portfolio. This must be 1 dimensional and must match the dimension of the data's
        last axis.

    freq: Frequency
        Frequency of the data. Can either be a string ('week', 'month', 'quarter', 'semi-annual', 'year') or
        an integer specifying the number of units per year. Week: 52, Month: 12, Quarter: 4, Semi-annual: 6,
        Year: 1.

    shock: float
        The amount to shock each asset class by. A positive number represents adding to the asset class by
        proportionately removing from the other asset class. A negative number represents removing from the
        asset class and adding to the other asset class proportionately.

    names: list of str
        Asset class names

    leveraged: bool
        If True, asset weights are allowed to go below 0 and above 1. This represents that the
        asset class can be shorted or levered.

    distribute: bool
        If True, asset value changes are distributed proportionately to all other asset classes. See Notes
        for more information.

    Returns
    -------
    Series
        A series with asset names as the index and annualized volatility as its value

    Examples
    --------
    >>> from perfana.datasets import load_cube
    >>> from perfana.monte_carlo import sensitivity_vol_m
    >>> data = load_cube()[..., :7]
    >>> weights = [0.25, 0.18, 0.13, 0.11, 0.24, 0.05, 0.04]
    >>> freq = 'quarterly'
    >>> shock = 0.05  # 5% absolute shock
    >>> sensitivity_vol_m(data, weights, freq, shock)
    Asset_1    0.113284
    Asset_2    0.121786
    Asset_3    0.113964
    Asset_4    0.109301
    Asset_5    0.104626
    Asset_6    0.106652
    Asset_7    0.106140
    Name: vol, dtype: float64
    """
    weight_matrix = _setup(weights, shock, leveraged, distribute)
    names = _setup_names(weights, names)
    vol = [volatility_m(cov_or_data, w, freq) for w in weight_matrix]

    return pd.Series(vol, names, name="vol")


def _setup(weights: Vector,
           shock: float = 0.05,
           leveraged=False,
           distribute=True):
    """Common setup for sensitivity analytics"""
    assert -1 <= shock <= 1, "shock must be between [-1, 1]"
    weights = np.ravel(weights)

    # if leverage is True:
    # Prevents shocks beyond the asset's current allocation. That is if the shock is
    # -5% and the asset only has 2% allocation, then the shock is effectively -2%.
    # And prevent shocks that bring the assets beyond 100% allocation.
    shocks = np.array([
        shock if (0 <= w + shock <= 1) or leveraged else
        -w if shock < 0 else 1 - w
        for w in weights
    ])

    n = len(weights)

    matrix = np.tile(weights, (n, 1)) + np.diag(shocks)
    if distribute:
        weight_matrix = np.tile(weights, (n, 1))
        np.fill_diagonal(weight_matrix, 0)

        matrix -= weight_matrix * (shocks / weight_matrix.sum(1))[:, None]

    return matrix


def _setup_names(weights: np.ndarray, names: List[str] = None):
    if names is None:
        return [f"Asset_{i + 1}" for i in range(len(weights))]

    names = list(names)
    assert len(names) == len(weights), "number of names given is not equal to length of weight vector"
    return names



from typing import Union

import pandas as pd
from pandas.api.extensions import register_series_accessor

from perfana.conversions import *
from perfana.core.utils import infer_frequency
from perfana.types import DateTimes


@register_series_accessor('pa')
class PaSeriesAccessor:
    def __init__(self, pd_obj: pd.Series):
        self._obj = pd_obj

    def to_prices(self, start: Union[float, int] = 100., log: bool = False) -> pd.Series:
        """
        Creates a price series from a returns series

        Parameters
        ----------
        start
            the starting prices

        log
            If data is a log returns data, set this to True

        Returns
        -------
        Series
            Asset price series
        """
        return to_prices(self._obj, start, log)

    def to_returns(self, log: bool = False) -> pd.Series:
        """
        Creates a returns series from price series

        Parameters
        ----------
        log
            If True, calculate the log returns. Otherwise, calculates the arithmetic returns

        Returns
        -------
        Series
            Asset returns series
        """
        return to_returns(self._obj, log)

    def to_time_series(self, dates: DateTimes) -> pd.Series:
        """
        Creates a time series.

        Parameters
        ----------
        dates
            A list of dates

        Returns
        -------
        Series
            A time series
        """
        return to_time_series(self._obj, dates)

    @property
    def frequency(self):
        """Frequency (periodicity) of the data"""
        return infer_frequency(self._obj, 'ignore')



import numpy as np
from copulae.core import cov2corr, is_psd

from perfana.types import Vector
from ._types import Attribution, Drawdown, Frequency, TailLoss
from ._utility import infer_frequency

__all__ = [
    "beta_m",
    "correlation_m",
    "cvar_attr",
    "cvar_m",
    "cvar_div_ratio",
    "diversification_m",
    "drawdown_m",
    "portfolio_cov",
    "prob_loss",
    "prob_under_perf",
    "tail_loss",
    "tracking_error_m",
    "vol_attr",
    "volatility_m",
]


def beta_m(cov_or_data: np.ndarray,
           weights: Vector,
           freq: Frequency = None,
           aid=0) -> float:
    """
    Derives the portfolio beta with respect to the specified asset class

    Notes
    -----
    The asset is identified by its index (aid) on the covariance matrix / simulated
    returns cube / weight vector. If a simulated returns data cube is given,
    the frequency of the data must be specified. In this case, the empirical covariance
    matrix would be used to derive the volatility.

    Parameters
    ----------
    cov_or_data
        Covariance matrix or simulated returns data cube.

    weights
        Weights of the portfolio. This must be 1 dimensional and must match the dimension of the
        covariance matrix shape or the simulated data's last axis.

    freq
        Frequency of the data. Can either be a string ('week', 'month', 'quarter', 'semi-annual', 'year') or
        an integer specifying the number of units per year. Week: 52, Month: 12, Quarter: 4, Semi-annual: 6,
        Year: 1.

    aid
        Asset index

    Returns
    -------
    float
        Portfolio beta with respect to asset class.

    Examples
    --------
    >>> from perfana.datasets import load_cube
    >>> from perfana.monte_carlo import portfolio_cov, beta_m
    >>> data = load_cube()[..., :3]  # first 3 asset classes only
    >>> weights = [0.33, 0.34, 0.33]
    >>> freq = 'quarterly'
    >>> dm_eq_id = 0  # calculate correlation with respect to developing markets equity
    >>> beta_m(data, weights, freq, dm_eq_id)
    1.3047194776321622
    """
    cov = _portfolio_volatility_structure(cov_or_data, weights, freq, aid)
    return cov[0, 1] / cov[1, 1]


def correlation_m(cov_or_data: np.ndarray,
                  weights: Vector,
                  freq: Frequency = None,
                  aid=0) -> float:
    """
    Derives the portfolio correlation with respect to the specified asset class

    Notes
    -----
    The asset is identified by its index (aid) on the covariance matrix / simulated
    returns cube / weight vector. If a simulated returns data cube is given,
    the frequency of the data must be specified. In this case, the empirical covariance
    matrix would be used to derive the volatility.

    Parameters
    ----------
    cov_or_data
        Covariance matrix or simulated returns data cube.

    weights
        Weights of the portfolio. This must be 1 dimensional and must match the dimension of the
        covariance matrix shape or the simulated data's last axis.

    freq
        Frequency of the data. Can either be a string ('week', 'month', 'quarter', 'semi-annual', 'year') or
        an integer specifying the number of units per year. Week: 52, Month: 12, Quarter: 4, Semi-annual: 6,
        Year: 1.

    aid
        Asset index

    Returns
    -------
    float
        Portfolio correlation with respect to asset class

    Examples
    --------
    >>> from perfana.datasets import load_cube
    >>> from perfana.monte_carlo import portfolio_cov, correlation_m
    >>> data = load_cube()[..., :3]  # first 3 asset classes only
    >>> weights = [0.33, 0.34, 0.33]
    >>> freq = 'quarterly'
    >>> dm_eq_id = 0  # calculate correlation with respect to developing markets equity
    >>> correlation_m(data, weights, freq, dm_eq_id)
    0.9642297301278216
    """
    cov = _portfolio_volatility_structure(cov_or_data, weights, freq, aid)
    return cov2corr(cov)[0, 1]


def cvar_attr(data: np.ndarray,
              weights: Vector,
              alpha=0.95,
              rebalance=True,
              invert=True) -> Attribution:
    """
    Calculates the CVaR (Expected Shortfall) attribution for each asset class in the portfolio.

    Notes
    -----
    From a mathematical point of view, the alpha value (confidence level for calculation)
    should be taken at the negative extreme of the distribution. However, the default is
    set to ease the practitioner.

    The return values are defined as follows:

    - **marginal**
        The absolute marginal contribution of the asset class towards the portfolio CVaR.
        It is essentially the percentage attribution multiplied by the portfolio CVaR.

    - **percentage**
        The percentage contribution of the asset class towards the portfolio CVaR. This number
        though named in percentage is actually in decimals. Thus 0.01 represents a 1% contribution.


    Parameters
    ----------
    data
        Monte carlo simulation data. This must be 3 dimensional with the axis representing time, trial
        and asset respectively.

    weights
        Weights of the portfolio. This must be 1 dimensional and must match the dimension of the data's
        last axis.

    alpha
        Confidence level for calculation.

    rebalance
        If True, portfolio is assumed to be rebalanced at every step.

    invert
        Whether to invert the confidence interval level. See Notes.

    Returns
    -------
    Attribution
        A named tuple of relative and absolute CVaR (expected shortfall) attribution respectively.
        The absolute attribution is the CVaR of the simulated data over time multiplied by the
        percentage attribution.

    Examples
    --------
    >>> from perfana.datasets import load_cube
    >>> from perfana.monte_carlo import cvar_attr
    >>> cube = load_cube()[..., :3]
    >>> weights = [0.33, 0.34, 0.33]
    >>> attr = cvar_attr(cube, weights, alpha=0.95)
    >>> attr.marginal
    array([-0.186001  , -0.35758411, -0.20281477])
    >>> attr.percentage
    array([0.24919752, 0.47907847, 0.27172401])
    >>> attr.marginal is attr[0]
    True
    >>> attr.percentage is attr[1]
    True
    """
    data = np.asarray(data)
    port_cvar = cvar_m(data, weights, alpha, rebalance, invert)

    assert isinstance(alpha, float) and 0 <= alpha <= 1, "alpha must be a float between 0 and 1"
    alpha = 1 - alpha if invert else alpha
    t, n, a = data.shape
    k = int(alpha * n)

    if rebalance:
        # calculate the CVaR quantile for each time period for the rebalanced portfolio
        port = data @ weights
        # calculate alpha quantile value at each time period
        quantiles = np.quantile(port, alpha, 1)

        cvar_decomp = np.zeros((t, a))  # cvar decomposition at each time period for each asset class
        for i, q, p, d in zip(range(t), quantiles, port, data):
            # for each time period, get the mean of the values that are less than the quantile for that
            # time period along the trials axis. Multiply resulting vector with the weights
            es = d[p <= q].mean(0) * weights
            cvar_decomp[i] = es / es.sum()

        attr = cvar_decomp.mean(0)
    else:
        data = (data + 1).prod(0) - 1  # portfolio drift

        # sort data according to the portfolio returns then truncate at the alpha value
        sorted_data = data[np.argsort(data @ weights)][:k]

        # take the mean along each trials axis multiplied by weight
        es = sorted_data.mean(0) * weights
        attr = es / es.sum()

    return Attribution(port_cvar * attr, attr)


def cvar_div_ratio(data: np.ndarray,
                   weights: Vector,
                   alpha=0.95,
                   rebalance=True,
                   invert=True) -> float:
    """
    Calculates the CVaR (Expected Shortfall) tail diversification ratio of the portfolio

    Notes
    -----
    From a mathematical point of view, the alpha value (confidence level for calculation)
    should be taken at the negative extreme of the distribution. However, the default is
    set to ease the practitioner.

    Parameters
    ----------
    data
        Monte carlo simulation data. This must be 3 dimensional with the axis representing time, trial
        and asset respectively.

    weights
        Weights of the portfolio. This must be 1 dimensional and must match the dimension of the data's
        last axis.

    alpha
        Confidence level for calculation.

    rebalance
        If True, portfolio is assumed to be rebalanced at every step.

    invert
        Whether to invert the confidence interval level. See Notes.

    Returns
    -------
    float
        CVaR (Expected Shortfall) tail diversification ratio

    Examples
    --------
    >>> from perfana.datasets import load_cube
    >>> from perfana.monte_carlo import cvar_div_ratio
    >>> cube = load_cube()[..., :3]
    >>> weights = [0.33, 0.34, 0.33]
    >>> cvar_div_ratio(cube, weights)
    0.8965390850633622
    """
    data = np.asarray(data)
    port_cvar = cvar_m(data, weights, alpha, rebalance, invert)

    assert isinstance(alpha, float) and 0 <= alpha <= 1, "alpha must be a float between 0 and 1"
    alpha = 1 - alpha if invert else alpha

    asset_cvar = np.zeros_like(weights)
    for i in range(len(weights)):
        d = (data[..., i] + 1).prod(0) - 1
        q = np.quantile(d, alpha)
        asset_cvar[i] = d[d <= q].mean()

    return (asset_cvar @ weights) / port_cvar


def cvar_m(data: np.ndarray,
           weights: Vector,
           alpha=0.95,
           rebalance=True,
           invert=True):
    """
    Calculates the Conditional Value at Risk (Expected Shortfall) of the portfolio.

    Notes
    -----
    From a mathematical point of view, the alpha value (confidence level for calculation)
    should be taken at the negative extreme of the distribution. However, the default is
    set to ease the practitioner.

    Parameters
    ----------
    data
        Monte carlo simulation data. This must be 3 dimensional with the axis representing time, trial
        and asset respectively.

    weights
        Weights of the portfolio. This must be 1 dimensional and must match the dimension of the data's
        last axis.

    alpha
        Confidence level for calculation.

    rebalance
        If True, portfolio is assumed to be rebalanced at every step.

    invert
        Whether to invert the confidence interval level. See Notes.

    Returns
    -------
    float
        CVaR (Expected Shortfall) of the portfolio

    Examples
    --------
    >>> from perfana.datasets import load_cube
    >>> from perfana.monte_carlo import cvar_m
    >>> cube = load_cube()[..., :3]
    >>> weights = [0.33, 0.34, 0.33]
    >>> cvar_m(cube, weights)
    -0.7463998716846179
    """

    data = np.asarray(data)
    assert isinstance(alpha, float) and 0 <= alpha <= 1, "alpha must be a float between 0 and 1"
    alpha = 1 - alpha if invert else alpha

    if rebalance:
        data = (data @ weights + 1).prod(0) - 1
    else:
        data = (data + 1).prod(0) @ weights - 1

    return data[data <= np.quantile(data, alpha)].mean()


def diversification_m(cov_or_data: np.ndarray,
                      weights: Vector,
                      freq: Frequency) -> float:
    """
    Derives the diversification ratio of the portfolio

    Parameters
    ----------
    cov_or_data
        Covariance matrix or simulated returns data cube.

    weights
        Weights of the portfolio. This must be 1 dimensional and must match the dimension of the data's
        last axis.

    freq
        Frequency of the data. Can either be a string ('week', 'month', 'quarter', 'semi-annual', 'year') or
        an integer specifying the number of units per year. Week: 52, Month: 12, Quarter: 4, Semi-annual: 6,
        Year: 1.

    Returns
    -------
    float
        Tracking error of the portfolio

    Examples
    --------
    >>> from perfana.datasets import load_cube
    >>> from perfana.monte_carlo import portfolio_cov, diversification_m
    >>> data = load_cube()[..., :7]  # first 7 asset classes
    >>> weights = [0.25, 0.18, 0.13, 0.11, 0.24, 0.05, 0.04]
    >>> freq = 'quarterly'
    >>> diversification_m(data, weights, freq)
    """
    cov = _get_covariance_matrix(cov_or_data, freq)
    vol = volatility_m(cov, weights)
    _, asset_vol = cov2corr(cov, True)

    return (weights * asset_vol).sum() / vol


def drawdown_m(data: np.ndarray,
               weights: Vector,
               geometric=True,
               rebalance=True) -> Drawdown:
    """
    Calculates the drawdown statistics

    Parameters
    ----------
    data
        Monte carlo simulation data. This must be 3 dimensional with the axis representing time, trial
        and asset respectively.

    weights
        Weights of the portfolio. This must be 1 dimensional and must match the dimension of the data's
        last axis.

    geometric
        If True, calculates the geometric mean, otherwise, calculates the arithmetic mean.

    rebalance
        If True, portfolio is assumed to be rebalanced at every step.

    Returns
    -------
    Drawdown
        A named tuple containing the average maximum drawdown and the drawdown path for each simulation
        instance.

    Examples
    --------
    >>> from perfana.datasets import load_cube
    >>> from perfana.monte_carlo import drawdown_m
    >>> data = load_cube()[..., :7]
    >>> weights = [0.25, 0.18, 0.13, 0.11, 0.24, 0.05, 0.04]
    >>> dd = drawdown_m(data, weights)
    >>> dd.average
    -0.3198714473717889
    >>> dd.paths.shape
    (80, 1000)
    """
    data = np.asarray(data)
    weights = np.ravel(weights)

    # cumulative returns per asset class
    if rebalance:
        if geometric:
            cum_ret = (data @ weights + 1).cumprod(0)
        else:
            cum_ret = (data @ weights).cumsum(0) + 1
    else:
        if geometric:
            cum_ret = ((data + 1).cumprod(0) - 1) @ weights + 1
        else:
            cum_ret = data.cumsum(1) @ weights + 1

    # get cumulative maximum along path
    cum_max = np.maximum.accumulate(np.vstack([np.ones(data.shape[1]), cum_ret]), 0)[1:]
    # drawdown path formula
    dd_paths = cum_ret / cum_max - 1
    average_max_dd = dd_paths.min(0).mean()

    return Drawdown(average_max_dd, dd_paths)


def portfolio_cov(data: np.ndarray,
                  freq: Frequency) -> np.ndarray:
    """
    Forms the empirical portfolio covariance matrix from the simulation data cube

    Parameters
    ----------
    data
        Monte carlo simulation data. This must be 3 dimensional with the axis representing time, trial
        and asset respectively.

    freq
        Frequency of the data. Can either be a string ('week', 'month', 'quarter', 'semi-annual', 'year') or
        an integer specifying the number of units per year. Week: 52, Month: 12, Quarter: 4, Semi-annual: 6,
        Year: 1.

    Returns
    -------
    array_like of float
        Empirical portfolio covariance matrix

    Examples
    --------
    >>> from perfana.datasets import load_cube
    >>> from perfana.monte_carlo import portfolio_cov
    >>> data = load_cube()[..., :3]  # first 3 asset classes only
    >>> portfolio_cov(data, 'quarterly').round(4)
    array([[0.0195, 0.0356, 0.021 ],
           [0.0356, 0.0808, 0.0407],
           [0.021 , 0.0407, 0.0239]])
    """
    freq = infer_frequency(freq)
    data = np.asarray(data)
    t, n, a = data.shape
    y = t // freq

    # convert to annual returns then calculate the average covariance across time
    ar = (data + 1).reshape(y, freq, n, a).prod(1) - 1
    return np.mean([np.cov(ar[i], rowvar=False) for i in range(y)], 0)


def prob_loss(data: np.ndarray,
              weights: Vector,
              rebalance=True,
              terminal=False) -> float:
    """
    Calculates the probability of the portfolio suffering a loss。

    Parameters
    ----------
    data
        Monte carlo simulation data. This must be 3 dimensional with the axis representing time, trial
        and asset respectively.

    weights
        Weights of the portfolio. This must be 1 dimensional and must match the dimension of the data's
        last axis.

    rebalance
        If True, portfolio is assumed to be rebalanced at every step.

    terminal
        If True, this only compares the probability of a loss at the last stage. If False (default),
        the calculation will take into account if the portfolio was "ruined" and count it as a loss
        even though the terminal value is positive.

    Returns
    -------
    float
        A named tuple containing the probability of underperformance and loss

    Examples
    --------
    >>> from perfana.datasets import load_cube
    >>> from perfana.monte_carlo import prob_loss
    >>> data = load_cube()
    >>> weights = [0.25, 0.18, 0.13, 0.11, 0.24, 0.05, 0.04]
    >>> prob_loss(data, weights)
    0.198
    """
    weights = np.ravel(weights)
    n = len(weights)

    port_data = data[..., :n]
    if rebalance:
        port_cum_ret = (port_data @ weights + 1).cumprod(0) - 1
    else:
        port_cum_ret = ((port_data + 1).cumprod(0) - 1) @ weights

    loss = np.asarray(port_cum_ret < 0)[-1]

    if terminal:
        return loss.mean()
    else:
        return ((port_cum_ret < -1).any(0) | loss).mean()


def prob_under_perf(data: np.ndarray,
                    weights: Vector,
                    bmk_weights: Vector,
                    rebalance=True,
                    terminal=False) -> float:
    """
    Calculates the probability of the portfolio underperforming the benchmark at the terminal state

    Parameters
    ----------
    data
        Monte carlo simulation data. This must be 3 dimensional with the axis representing time, trial
        and asset respectively.

    weights
        Weights of the portfolio. This must be 1 dimensional and must match the dimension of the data's
        last axis.

    bmk_weights
        Weights of the benchmark portfolio.

    rebalance
        If True, portfolio is assumed to be rebalanced at every step.

    terminal
        If True, this only compares the probability of underperformance at the last stage.
        If False (default), the calculation will take into account if the portfolio was "ruined"
        and count it as an underperformance against the benchmark even though the terminal value
        is higher than the benchmark. If both portfolios are "ruined", then it underperforms if
        it is ruined earlier.

    Returns
    -------
    float
        A named tuple containing the probability of underperformance and loss

    Examples
    --------
    >>> from perfana.datasets import load_cube
    >>> from perfana.monte_carlo import prob_under_perf
    >>> data = load_cube()
    >>> weights = [0.25, 0.18, 0.13, 0.11, 0.24, 0.05, 0.04]
    >>> bmk_weights = [0.65, 0.35]
    >>> prob_under_perf(data, weights, bmk_weights)
    0.863
    """
    weights, bmk_weights = np.ravel(weights), np.ravel(bmk_weights)
    n = len(weights)

    port_data, bmk_data = data[..., :n], data[..., n:]

    if rebalance:
        port_cum_ret = (port_data @ weights + 1).cumprod(0) - 1
        bmk_cum_ret = (bmk_data @ bmk_weights + 1).cumprod(0) - 1
    else:
        port_cum_ret = ((port_data + 1).cumprod(0) - 1) @ weights
        bmk_cum_ret = ((bmk_data + 1).cumprod(0) - 1) @ bmk_weights

    under_perf = np.asarray(port_cum_ret[-1] < bmk_cum_ret[-1])  # wrapped for autocomplete
    if terminal:
        return under_perf.mean()

    m = data.shape[1]
    for i, prow, brow in zip(range(m), port_cum_ret.T, bmk_cum_ret.T):
        for p, b in zip(prow, brow):
            if b <= -1:
                # can ignore the fact that p <= -1, cause there would be no underperformance either way
                under_perf[i] = False
                break
            elif p <= -1:
                under_perf[i] = True
                break

    return under_perf.mean()


def tail_loss(data: np.ndarray,
              weights: Vector,
              threshold=-0.3,
              rebalance=True) -> TailLoss:
    """
    Calculates the probability and expectation of a tail loss beyond a threshold

    Threshold by default is set at -0.3, which means find the probability that the portfolio loses more than
    30% of its value and the expected loss.

    Notes
    -----
    The return values are defined as follows:

    - **prob**
        Probability of having a tail loss exceeding the threshold

    - **expected_loss**
        Value of the expected loss for the portfolio at the threshold

    Parameters
    ----------
    data
        Monte carlo simulation data. This must be 3 dimensional with the axis representing time, trial
        and asset respectively.

    weights
        Weights of the portfolio. This must be 1 dimensional and must match the dimension of the data's
        last axis.

    threshold
        Portfolio loss threshold.

    rebalance
        If True, portfolio is assumed to be rebalanced at every step.

    Returns
    -------
    TailLoss
        A named tuple containing the probability and expected loss of the portfolio exceeding the threshold.

    Examples
    --------
    >>> from perfana.datasets import load_cube
    >>> from perfana.monte_carlo import tail_loss
    >>> data = load_cube()[..., :3]  # first 3 asset classes only
    >>> weights = [0.33, 0.34, 0.33]
    >>> loss = tail_loss(data, weights, -0.3)
    >>> loss.prob
    0.241
    >>> loss.expected_loss
    -0.3978210273894446
    """
    data = np.asarray(data)
    weights = np.ravel(weights)

    if rebalance:
        port = (data @ weights + 1).prod(0) - 1
    else:
        port = ((data + 1).prod(0) - 1) @ weights

    mask = port <= threshold
    prob = mask.mean()
    exp = 0 if sum(mask) == 0 else port[mask].mean()

    return TailLoss(prob, exp)


def tracking_error_m(cov_or_data: np.ndarray,
                     weights: Vector,
                     bmk_weights: Vector,
                     freq: Frequency) -> float:
    """
    Calculates the tracking error with respect to the benchmark.

    If a covariance matrix is used as the data, the benchmark components must be placed after the
    portfolio components. If a simulated returns cube is used as the data, the benchmark components
    must be placed after the portfolio components.

    Parameters
    ----------
    cov_or_data
        Covariance matrix or simulated returns data cube.

    weights
        Weights of the portfolio. This must be 1 dimensional and must match the dimension of the data's
        last axis.

    bmk_weights
        Weights of the benchmark portfolio.

    freq
        Frequency of the data. Can either be a string ('week', 'month', 'quarter', 'semi-annual', 'year') or
        an integer specifying the number of units per year. Week: 52, Month: 12, Quarter: 4, Semi-annual: 6,
        Year: 1.

    Returns
    -------
    float:
        Tracking error of the portfolio

    Examples
    --------
    >>> from perfana.datasets import load_cube
    >>> from perfana.monte_carlo import portfolio_cov, tracking_error_m
    >>> data = load_cube()
    >>> weights = [0.25, 0.18, 0.13, 0.11, 0.24, 0.05, 0.04]
    >>> bmk_weights = [0.65, 0.35]
    >>> freq = 'quarterly'
    >>> tracking_error_m(data, weights, bmk_weights, freq)
    0.031183281273726802
    """
    weights = np.ravel(weights)
    bmk_weights = np.ravel(bmk_weights)
    w = np.hstack((weights, -bmk_weights))

    cov = _get_covariance_matrix(cov_or_data, freq)
    return (w @ cov @ w.T) ** 0.5


def vol_attr(cov_or_data: np.ndarray,
             weights: Vector,
             freq: Frequency) -> Attribution:
    """
    Derives the volatility attribution given a data cube and weights.

    Notes
    -----
    The return values are defined as follows:

    **marginal**
        The absolute marginal contribution of the asset class towards the portfolio volatility.
        It is essentially the percentage attribution multiplied by the portfolio volatility.

    **percentage**
        The percentage contribution of the asset class towards the portfolio volatility. This number
        though named in percentage is actually in decimals. Thus 0.01 represents a 1% contribution.

    Parameters
    ----------
    cov_or_data
        Covariance matrix or simulated returns data cube.

    weights
        Weights of the portfolio. This must be 1 dimensional and must match the dimension of the data's
        last axis.

    freq
        Frequency of the data. Can either be a string ('week', 'month', 'quarter', 'semi-annual', 'year') or
        an integer specifying the number of units per year. Week: 52, Month: 12, Quarter: 4, Semi-annual: 6,
        Year: 1.

    Returns
    -------
    Attribution
        A named tuple of relative and absolute volatility attribution respectively. The absolute attribution
        is the volatility of the simulated data over time multiplied by the percentage attribution.

    Examples
    --------
    >>> from perfana.datasets import load_cube
    >>> from perfana.monte_carlo import vol_attr
    >>> data = load_cube()[..., :3]  # first 3 asset classes only
    >>> weights = [0.33, 0.34, 0.33]
    >>> freq = 'quarterly'
    >>> attr = vol_attr(data, weights, freq)
    >>> attr.marginal.round(4)
    array([0.2352, 0.5006, 0.2643])
    >>> attr.percentage.round(4)
    array([0.0445, 0.0947, 0.05  ])
    >>> attr.marginal is attr[0]
    True
    >>> attr.percentage is attr[1]
    True
    """
    weights = np.ravel(weights)
    cov = _get_covariance_matrix(cov_or_data, freq)
    vol = volatility_m(cov, weights)
    attr = (weights * (cov @ weights)) / (weights @ cov @ weights.T)

    return Attribution(vol * attr, attr)


def volatility_m(cov_or_data: np.ndarray,
                 weights: Vector,
                 freq: Frequency = None) -> float:
    """
    Calculates the portfolio volatility given a simulated returns cube or a covariance matrix

    Notes
    -----
    If a simulated returns data cube is given, the frequency of the data must be specified.
    In this case, the empirical covariance matrix would be used to derive the volatility.

    Parameters
    ----------
    cov_or_data
        Covariance matrix or simulated returns data cube.

    weights
        Weights of the portfolio. This must be 1 dimensional and must match the dimension of the
        covariance matrix shape or the simulated data's last axis.

    freq
        Frequency of the data. Can either be a string ('week', 'month', 'quarter', 'semi-annual', 'year') or
        an integer specifying the number of units per year. Week: 52, Month: 12, Quarter: 4, Semi-annual: 6,
        Year: 1.

    Returns
    -------
    float
        Portfolio volatility.

    Examples
    --------
    >>> from perfana.datasets import load_cube
    >>> from perfana.monte_carlo import portfolio_cov, volatility_m
    >>> data = load_cube()[..., :3]  # first 3 asset classes only
    >>> weights = [0.33, 0.34, 0.33]
    >>> freq = 'quarterly'
    >>> cov_mat = portfolio_cov(data, freq).round(4)  # empirical covariance matrix
    >>> # Using covariance matrix
    >>> volatility_m(cov_mat, weights)
    0.1891091219375734
    >>> # Using the simulated returns data cube
    >>> volatility_m(data, weights, freq)
    0.1891091219375734
    """
    s = _portfolio_volatility_structure(cov_or_data, weights, freq)
    return s[0, 0] ** 0.5


def _get_covariance_matrix(cov_or_data: np.ndarray, freq: Frequency):
    """
    Helper to derive a covariance matrix.

    If simulated cube is put in, derive empirical covariance matrix. Otherwise, check that matrix
    is a covariance matrix and return it.
    """

    assert cov_or_data.ndim in (2, 3), "only 2D covariance matrix or a 3D simulation returns tensor is allowed"
    cov_or_data = np.asarray(cov_or_data)
    n = cov_or_data.shape[-1]  # number of assets must be last dimension of cube or covariance matrix

    if cov_or_data.ndim == 3:
        # if 3D, get empirical covariance matrix
        assert n == cov_or_data.shape[2], "number of weights does not match simulation cube asset quantity"
        assert freq is not None, "frequency must be specified if deriving covariance structure from simulated data"

        cov_or_data = portfolio_cov(cov_or_data, freq)

    assert is_psd(cov_or_data), "covariance matrix is not positive semi-definite"
    return cov_or_data  # this will be a covariance matrix


def _portfolio_volatility_structure(cov_or_data: np.ndarray,
                                    weights: Vector,
                                    freq: Frequency = None,
                                    aid: int = 0) -> np.ndarray:
    """
    Derive the portfolio volatility structure

    Parameters
    ----------
    cov_or_data
        Covariance matrix or simulated returns data cube.

    weights
        Weights of the portfolio. This must be 1 dimensional and must match the dimension of the
        covariance matrix shape or the simulated data's last axis.

    freq
        Frequency of the data. Can either be a string ('week', 'month', 'quarter', 'semi-annual', 'year') or
        an integer specifying the number of units per year. Week: 52, Month: 12, Quarter: 4, Semi-annual: 6,
        Year: 1.

    aid
        Asset index

    Returns
    -------
    covariance matrix
        Covariance matrix of portfolio volatility structure
    """
    cov_or_data = np.asarray(cov_or_data)
    weights = np.ravel(weights)
    n = len(weights)

    cov = _get_covariance_matrix(cov_or_data, freq)
    w = np.vstack([weights, np.zeros(n)])

    assert 0 <= aid <= len(w), f"asset id must be between [0, {len(w)}]"
    w[1, aid] = 1
    return w @ cov @ w.T



import numpy as np
from copulae.core import is_psd

from perfana.monte_carlo.returns import annualized_bmk_returns_m, annualized_returns_m
from perfana.monte_carlo.risk import tracking_error_m, volatility_m
from perfana.types import Vector
from ._types import Frequency

__all__ = ["sharpe_ratio_m", "sharpe_ratio_bmk_m"]


def sharpe_ratio_m(data: np.ndarray,
                   weights: Vector,
                   freq: Frequency,
                   cov: np.ndarray = None,
                   geometric: bool = True,
                   rebalance: bool = True) -> float:
    """
    Calculates the Sharpe Ratio of portfolio

    Parameters
    ----------
    data
        Monte carlo simulation data. This must be 3 dimensional with the axis representing time, trial
        and asset respectively.

    weights
        Weights of the portfolio. This must be 1 dimensional and must match the dimension of the data's
        last axis.

    freq
        Frequency of the data. Can either be a string ('week', 'month', 'quarter', 'semi-annual', 'year') or
        an integer specifying the number of units per year. Week: 52, Month: 12, Quarter: 4, Semi-annual: 6,
        Year: 1.

    cov
        Covariance matrix. If not specified, the empirical covariance matrix from the joined data and
        benchmark data will be derived.

    geometric
        If True, calculates the geometric mean, otherwise, calculates the arithmetic mean.

    rebalance
        If True, portfolio is assumed to be rebalanced at every step.

    Returns
    -------
    float
        Sharpe ratio of the portfolio

    Examples
    --------
    >>> from perfana.datasets import load_cube
    >>> from perfana.monte_carlo import sharpe_ratio_m
    >>> cube = load_cube()[..., :3]
    >>> weights = [0.33, 0.34, 0.33]
    >>> freq = 'quarterly'
    >>> sharpe_ratio_m(cube, weights, freq)
    0.14242462120295252
    """
    gr = annualized_returns_m(data, weights, freq, geometric, rebalance)
    vol = volatility_m(cov if cov is not None else data, weights, freq)
    return gr / vol


def sharpe_ratio_bmk_m(data: np.ndarray,
                       weights: Vector,
                       bmk_weights: Vector,
                       freq: Frequency,
                       cov: np.ndarray = None,
                       geometric: bool = True,
                       rebalance: bool = True):
    """
    Calculates the Sharpe Ratio of the portfolio over the benchmark.

    The benchmark components must be placed after the portfolio components in the simulated returns cube.

    Parameters
    ----------
    data
        Monte carlo simulation data. This must be 3 dimensional with the axis representing time, trial
        and asset respectively.

    weights
        Weights of the portfolio. This must be 1 dimensional and must match the dimension of the data's
        last axis.

    bmk_weights
        Weights of the benchmark portfolio.

    freq
        Frequency of the data. Can either be a string ('week', 'month', 'quarter', 'semi-annual', 'year') or
        an integer specifying the number of units per year. Week: 52, Month: 12, Quarter: 4, Semi-annual: 6,
        Year: 1.

    cov
        Covariance matrix. If not specified, the empirical covariance matrix from the joined data and
        benchmark data will be derived.

    geometric
        If True, calculates the geometric mean, otherwise, calculates the arithmetic mean.

    rebalance
        If True, portfolio is assumed to be rebalanced at every step.

    Returns
    -------
    float
        Sharpe ratio of the portfolio over the benchmark

    Examples
    --------
    >>> from perfana.datasets import load_cube
    >>> from perfana.monte_carlo import sharpe_ratio_bmk_m
    >>> cube = load_cube()
    >>> weights = [0.25, 0.18, 0.13, 0.11, 0.24, 0.05, 0.04]
    >>> bmk_weights = [0.65, 0.35]
    >>> freq = 'quarterly'
    >>> sharpe_ratio_bmk_m(cube, weights, bmk_weights, freq)
    -0.2186945589389277
    """
    gr = annualized_bmk_returns_m(data, weights, bmk_weights, freq, geometric, rebalance)

    if cov is not None:
        n = len(weights) + len(bmk_weights)
        assert cov.shape == (n, n), "covariance matrix shape incorrect"
        assert is_psd(cov), "covariance matrix is not positive semi-definite"
        data = cov

    te = tracking_error_m(data, weights, bmk_weights, freq)
    return gr / te



from typing import Iterable, Optional, Union

import numpy as np

from perfana.types import Scalar, Vector, is_iterable
from perfana.utils import lru_cache_ext
from ._types import Attribution, Frequency
from ._utility import infer_frequency

__all__ = [
    "annualized_returns_m",
    "annualized_bmk_returns_m",
    "annualized_quantile_returns_m",
    "annualized_bmk_quantile_returns_m",
    "returns_attr",
    "returns_distribution",
    "returns_path",
]


def annualized_returns_m(data: np.ndarray,
                         weights: Vector,
                         freq: Frequency,
                         geometric: bool = True,
                         rebalance: bool = True) -> float:
    r"""
    Calculates the annualized returns from the Monte Carlo simulation

    The formula for annualized geometric returns is formulated by raising the compound return to the number of
    periods in a year, and taking the root to the number of total observations. For the rebalance, geometric
    returns, the annualized returns is derived by:

    .. math::

        &y = M / s \\
        &\frac{1}{N}\sum^N_i \left[\prod_j^T \left(1 + \sum^A_k (r_{ijk} \cdot w_k \right) \right]^{\frac{1}{y}} - 1

    where `s` is the number of observations in a year, and `M` is the total number of observations, `N` is
    the number of trials in the simulation, `T` is the number of trials in the simulation and `A` is the
    number of assets in the simulation.

    For simple returns (geometric=FALSE), the formula for the rebalanced case is:

    .. math::

        \frac{s}{NM} \left[\sum^N_i \sum^T_j \sum^A_k (r_{ijk} \cdot w_k) \right]

    Parameters
    ----------
    data
        Monte carlo simulation data. This must be 3 dimensional with the axis representing time, trial
        and asset respectively.

    weights
        Weights of the portfolio. This must be 1 dimensional and must match the dimension of the data's
        last axis.

    freq
        Frequency of the data. Can either be a string ('week', 'month', 'quarter', 'semi-annual', 'year') or
        an integer specifying the number of units per year. Week: 52, Month: 12, Quarter: 4, Semi-annual: 6,
        Year: 1.

    geometric
        If True, calculates the geometric mean, otherwise, calculates the arithmetic mean.

    rebalance
        If True, portfolio is assumed to be rebalanced at every step.

    Returns
    -------
    float
        Annualized returns of the portfolio

    Examples
    --------
    >>> from perfana.datasets import load_cube
    >>> from perfana.monte_carlo import annualized_returns_m
    >>> cube = load_cube()[..., :7]
    >>> weights = [0.25, 0.18, 0.13, 0.11, 0.24, 0.05, 0.04]
    >>> annualized_returns_m(cube, weights, 'month')
    0.02111728739277985
    """
    d = returns_distribution(data, weights, freq, geometric=geometric, rebalance=rebalance)
    return d.mean()


def annualized_bmk_returns_m(data: np.ndarray,
                             weights: Vector,
                             bmk_weights: Vector,
                             freq: Frequency,
                             geometric: bool = True,
                             rebalance: bool = True) -> float:
    """
    Calculates the returns of the portfolio relative to a benchmark portfolio.

    The benchmark components must be placed after the portfolio components in the simulated returns cube.

    Parameters
    ----------
    data
        Monte carlo simulation data. This must be 3 dimensional with the axis representing time, trial
        and asset respectively.

    weights
        Weights of the portfolio. This must be 1 dimensional and must match the dimension of the data's
        last axis.

    bmk_weights
        Weights of the benchmark portfolio.

    freq
        Frequency of the data. Can either be a string ('week', 'month', 'quarter', 'semi-annual', 'year') or
        an integer specifying the number of units per year. Week: 52, Month: 12, Quarter: 4, Semi-annual: 6,
        Year: 1.

    geometric
        If True, calculates the geometric mean, otherwise, calculates the arithmetic mean.

    rebalance
        If True, portfolio is assumed to be rebalanced at every step.

    Returns
    -------
    float
        The portfolio returns relative to the benchmark

    Examples
    --------
    >>> from perfana.datasets import load_cube
    >>> from perfana.monte_carlo import annualized_bmk_returns_m
    >>> cube = load_cube()
    >>> weights = [0.25, 0.18, 0.13, 0.11, 0.24, 0.05, 0.04]
    >>> bmk_weights = [0.65, 0.35]
    >>> freq = "quarterly"
    >>> annualized_bmk_returns_m(cube, weights, bmk_weights, freq)
    -0.006819613944426206
    """
    freq = infer_frequency(freq)
    w1, w2 = np.ravel(weights), np.ravel(bmk_weights)
    y = len(data) / freq  # number of years
    n = len(w1)
    port, bmk = data[..., :n], data[..., n:]

    if rebalance:
        if geometric:
            d = ((port @ w1) + 1).prod(0) / ((bmk @ w2) + 1).prod(0)
            return (np.sign(d) * np.abs(d) ** (1 / y)).mean() - 1
        else:
            return ((port @ w1) / (bmk @ w2)).mean() * freq
    else:
        d = ((port + 1).prod(0) @ w1) / ((bmk + 1).prod(0) @ w2)
        if geometric:
            return (np.sign(d) * np.abs(d) ** (1 / y)).mean() - 1
        else:
            return (d - 1).mean() * freq


def annualized_quantile_returns_m(data: np.ndarray,
                                  weights: Vector,
                                  quantile: Union[float, Iterable[float]],
                                  freq: Frequency,
                                  geometric: bool = True,
                                  rebalance: bool = True,
                                  interpolation="midpoint") -> Union[float, np.ndarray]:
    """
    Compute the q-th quantile of the returns in the simulated data cube.

    Parameters
    ----------
    data
        Monte carlo simulation data. This must be 3 dimensional with the axis representing time, trial
        and asset respectively.

    weights
        Weights of the portfolio. This must be 1 dimensional and must match the dimension of the data's
        last axis.

    quantile
        Quantile or sequence of quantiles to compute, which must be between 0 and 1 inclusive

    freq
        Frequency of the data. Can either be a string ('week', 'month', 'quarter', 'semi-annual', 'year') or
        an integer specifying the number of units per year. Week: 52, Month: 12, Quarter: 4, Semi-annual: 6,
        Year: 1.

    geometric
        If True, calculates the geometric mean, otherwise, calculates the arithmetic mean.

    rebalance
        If True, portfolio is assumed to be rebalanced at every step.

    interpolation
        This optional parameter specifies the interpolation method to use when the desired quantile
        lies between two data points ``i < j``:

            * linear: ``i + (j - i) * fraction``, where ``fraction``
              is the fractional part of the index surrounded by ``i``
              and ``j``.
            * lower: ``i``.
            * higher: ``j``.
            * nearest: ``i`` or ``j``, whichever is nearest.
            * midpoint: ``(i + j) / 2``.
    Returns
    -------
    float
        The returns of the portfolio relative to the benchmark at the specified quantile

    Examples
    --------
    >>> from perfana.datasets import load_cube
    >>> from perfana.monte_carlo import annualized_quantile_returns_m
    >>> cube = load_cube()[..., :7]
    >>> weights = [0.25, 0.18, 0.13, 0.11, 0.24, 0.05, 0.04]
    >>> freq = "quarterly"
    >>> q = 0.25
    >>> annualized_quantile_returns_m(cube, weights, q, freq)
    0.005468353416130167
    >>> q = [0.25, 0.75]
    >>> annualized_quantile_returns_m(cube, weights, q, freq)
    array([0.00546835, 0.03845033])
    """
    w = np.ravel(weights)  # force weights to be a 1D vector
    d = returns_distribution(data, w, freq, True, geometric, rebalance)
    return np.quantile(d, quantile, interpolation=interpolation)


def annualized_bmk_quantile_returns_m(data: np.ndarray,
                                      weights: Vector,
                                      bmk_weights: Vector,
                                      quantile: Union[float, Iterable[float]],
                                      freq: Frequency,
                                      geometric: bool = True,
                                      rebalance: bool = True,
                                      interpolation="midpoint") -> Union[float, np.ndarray]:
    """
    Compares the annualized returns against a benchmark at the specified quantiles.

    The benchmark components must be placed after the portfolio components in the simulated returns cube.

    Parameters
    ----------
    data
        Monte carlo simulation data. This must be 3 dimensional with the axis representing time, trial
        and asset respectively.

    weights
        Weights of the portfolio. This must be 1 dimensional and must match the dimension of the data's
        last axis.

    bmk_weights
        Weights of the benchmark portfolio.

    quantile
        Quantile or sequence of quantiles to compute, which must be between 0 and 1 inclusive

    freq
        Frequency of the data. Can either be a string ('week', 'month', 'quarter', 'semi-annual', 'year') or
        an integer specifying the number of units per year. Week: 52, Month: 12, Quarter: 4, Semi-annual: 6,
        Year: 1.

    geometric
        If True, calculates the geometric mean, otherwise, calculates the arithmetic mean.

    rebalance
        If True, portfolio is assumed to be rebalanced at every step.

    interpolation
        This optional parameter specifies the interpolation method to use when the desired quantile
        lies between two data points ``i < j``:

            * linear: ``i + (j - i) * fraction``, where ``fraction``
              is the fractional part of the index surrounded by ``i``
              and ``j``.
            * lower: ``i``.
            * higher: ``j``.
            * nearest: ``i`` or ``j``, whichever is nearest.
            * midpoint: ``(i + j) / 2``.

    Returns
    -------
    float or array_like of floats
        The returns of the portfolio over the benchmark at the specified quantiles

    Examples
    --------
    >>> from perfana.datasets import load_cube
    >>> from perfana.monte_carlo import annualized_bmk_quantile_returns_m
    >>> cube = load_cube()
    >>> weights = [0.25, 0.18, 0.13, 0.11, 0.24, 0.05, 0.04]
    >>> bmk_weights = [0.65, 0.35]
    >>> freq = "quarterly"
    >>> q = 0.25
    >>> annualized_bmk_quantile_returns_m(cube, weights, bmk_weights, q, freq)
    -0.010792419409674459
    >>> q = [0.25, 0.75]
    >>> annualized_bmk_quantile_returns_m(cube, weights, bmk_weights, q, freq)
    array([-0.01079242, -0.0025487 ])
    """
    freq = infer_frequency(freq)
    w1, w2 = np.ravel(weights), np.ravel(bmk_weights)
    y = len(data) / freq  # number of years
    n = len(w1)
    port, bmk = data[..., :n], data[..., n:]

    if rebalance:
        if geometric:
            d = ((port @ w1) + 1).prod(0) / ((bmk @ w2) + 1).prod(0)
            return np.quantile((np.sign(d) * np.abs(d) ** (1 / y)), quantile, interpolation=interpolation) - 1
        else:
            return np.quantile(((port @ w1) / (bmk @ w2)), quantile, interpolation=interpolation) * freq
    else:
        d = ((port + 1).prod(0) @ w1) / ((bmk + 1).prod(0) @ w2)
        if geometric:
            return np.quantile((np.sign(d) * np.abs(d) ** (1 / y)), quantile, interpolation=interpolation) - 1
        else:
            return np.quantile((d - 1), quantile, interpolation=interpolation) * freq


def returns_attr(data: np.ndarray,
                 weights: Vector,
                 freq: Frequency,
                 geometric: bool = True,
                 rebalance: bool = True) -> Attribution:
    """
    Derives the returns attribution given a data cube and weights.

    Notes
    -----
    The return values are defined as follows:

    - **marginal**
        The absolute marginal contribution of the asset class towards the portfolio returns.
        It is essentially the percentage attribution multiplied by the portfolio returns.

    - **percentage**
        The percentage contribution of the asset class towards the portfolio returns. This number
        though named in percentage is actually in decimals. Thus 0.01 represents a 1% contribution.

    Parameters
    ----------
    data
        Monte carlo simulation data. This must be 3 dimensional with the axis representing time, trial
        and asset respectively.

    weights
        Weights of the portfolio. This must be 1 dimensional and must match the dimension of the data's
        last axis.

    freq
        Frequency of the data. Can either be a string ('week', 'month', 'quarter', 'semi-annual', 'year') or
        an integer specifying the number of units per year. Week: 52, Month: 12, Quarter: 4, Semi-annual: 6,
        Year: 1.

    geometric
        If True, calculates the geometric mean, otherwise, calculates the arithmetic mean.

    rebalance
        If True, portfolio is assumed to be rebalanced at every step.

    Returns
    -------
    Attribution
        A named tuple of marginal and percentage returns attribution respectively. The marginal attribution
        is the returns of the simulated data over time multiplied by the percentage attribution.

    Examples
    --------
    >>> from perfana.datasets import load_cube
    >>> from perfana.monte_carlo import returns_attr
    >>> cube = load_cube()[..., :3]
    >>> weights = [0.33, 0.34, 0.33]
    >>> freq = "quarterly"
    >>> attr = returns_attr(cube, weights, freq)
    >>> attr.marginal
    array([0.00996204, 0.00733369, 0.00963802])
    >>> attr.percentage
    array([0.36987203, 0.27228623, 0.35784174])
    >>> attr.marginal is attr[0]
    True
    >>> attr.percentage is attr[1]
    True
    """

    data = np.asarray(data)
    t, n, a = data.shape
    assert a == len(weights), "length of weights must match number of assets in data cube"

    freq = infer_frequency(freq)

    # calculate annualized returns before reshape
    gr = annualized_returns_m(data, weights, freq, geometric, rebalance)
    data = (data + 1).reshape(t // freq, freq, n, a).prod(1) - 1
    ar = data.mean((0, 1))  # arithmetic returns

    # TODO Check if non-rebalanced attribution are still calculated similarly
    attr_p = (ar * weights) / (ar @ weights)
    attr_m = attr_p * gr

    return Attribution(attr_m, attr_p)


@lru_cache_ext
def returns_distribution(data: np.ndarray,
                         weights: Vector,
                         freq: Optional[Frequency] = None,
                         annualize=True,
                         geometric: bool = True,
                         rebalance: bool = True) -> np.ndarray:
    r"""
    Calculates the returns distribution of the simulation cube

    Parameters
    ----------
    data
        Monte carlo simulation data. This must be 3 dimensional with the axis representing time, trial
        and asset respectively.

    weights
        Weights of the portfolio. This must be 1 dimensional and must match the dimension of the data's
        last axis.

    freq
        Frequency of the data. Can either be a string ('week', 'month', 'quarter', 'semi-annual', 'year') or
        an integer specifying the number of units per year. Week: 52, Month: 12, Quarter: 4, Semi-annual: 6,
        Year: 1. If :code:`annualize` is False, :code:`freq` can be ignored.

    annualize
        If true, the returns distribution values are annualized

    geometric
        If True, calculates the geometric mean, otherwise, calculates the arithmetic mean.

    rebalance
        If True, portfolio is assumed to be rebalanced at every step.

    Returns
    -------
    Array
        A vector of the distribution of returns

    Examples
    --------
    >>> from perfana.datasets import load_cube
    >>> from perfana.monte_carlo import returns_distribution
    >>> cube = load_cube()[..., :3]
    >>> weights = [0.33, 0.34, 0.33]
    >>> freq = "quarterly"
    >>> returns_distribution(cube, weights, freq).shape
    (1000,)
    """
    w = np.ravel(weights)  # force weights to be a 1D vector

    if annualize:
        freq = infer_frequency(freq)
        y = len(data) / freq  # number of years
    else:
        y = None

    if rebalance:
        if geometric:
            d = (data @ w + 1).prod(0)
            if annualize:
                d = np.sign(d) * np.abs(d) ** (1 / y) - 1
            else:
                d -= 1

        else:
            d = (data @ w).mean(0)
            if annualize:
                d *= freq
    else:
        d = ((data + 1).prod(0) @ w)
        if geometric:
            if annualize:
                d = (np.sign(d) * np.abs(d) ** (1 / y)) - 1
            else:
                d -= 1
        else:
            d = (d - 1).mean(0)
            if annualize:
                d = d * freq

    return d


def returns_path(data: np.ndarray,
                 weights: Vector,
                 rebalance: bool = True,
                 quantile: Union[Scalar, Vector] = None) -> np.ndarray:
    r"""
    Returns a matrix of the returns path of the portfolio.

    The first axis represents the time and the second axis represents the trials.

    If the :code:`quantile` argument is specified, the specific quantile for each time period will be
    returned. Thus, if the 0.75 quantile is specified, it is the 75th quantile for each time period and
    not the path the 75th quantile in the terminal period took.

    Parameters
    ----------
    data
        Monte carlo simulation data. This must be 3 dimensional with the axis representing time, trial
        and asset respectively.

    weights
        Weights of the portfolio. This must be 1 dimensional and must match the dimension of the data's
        last axis.

    rebalance
        If True, portfolio is assumed to be rebalanced at every step.

    quantile
        Quantile or sequence of quantiles to compute, which must be between 0 and 1 inclusive

    Returns
    -------
    ndarray
        The returns path for the portfolio

    Examples
    --------
    >>> from perfana.datasets import load_cube
    >>> from perfana.monte_carlo import returns_path
    >>> cube = load_cube()[..., :3]
    >>> weights = [0.33, 0.34, 0.33]
    >>> returns_path(cube, weights).shape
    (1000, 81)
    >>> returns_path(cube, weights, quantile=0.75).shape  # 75th quantile
    (81, )
    >>> returns_path(cube, weights, quantile=[0.25, 0.5, 0.75]).shape  # 25th, 50th and 75th quantile
    (3, 81)

    .. plot:: plots/returns_path.py
        :include-source:
    """
    w = np.ravel(weights)  # force weights to be a 1D vector

    # add zeros to the start
    _, n, a = data.shape
    data = np.append(np.zeros([1, n, a]), data, 0)

    if rebalance:
        d = (data @ w + 1).cumprod(0)
    else:
        d = (data + 1).cumprod(0) @ w

    if quantile is None:
        return d.T

    if not is_iterable(quantile):
        quantile = [quantile]

    quantile = np.asarray(quantile)
    assert np.alltrue([0 <= quantile, quantile <= 1]), "quantile values must be between 0 and 1"

    return np.squeeze(np.quantile(d, quantile, 1))



from collections import abc
from typing import Optional, Union

import numpy as np
import pandas as pd

from perfana.types import DateTimes, TimeSeriesData, Vector

__all__ = ['to_prices', 'to_returns', 'to_time_series']


def to_prices(returns: TimeSeriesData, start: Union[Vector, float, int] = 100, log: bool = False) -> TimeSeriesData:
    """
    Calculates the price index given a series of returns

    Parameters
    ----------
    returns
        Price series

    start
        The starting prices

    log
        If series is a log returns series, set this to True

    Returns
    -------
    TimeSeriesData
        Price index series
    """
    returns = to_time_series(returns)

    if isinstance(start, abc.Iterable):
        if not isinstance(returns, pd.DataFrame):
            raise TypeError("<returns> data must be a DataFrame if <start> is a vector")

        start = np.asarray(start)
        if start.size == 1:
            start = float(start)
        elif len(start) != len(returns.columns):
            raise ValueError('<start> vector length not equal to the number of columns in <returns> data')

    returns.replace(np.nan, 0, inplace=True)
    if log:
        return np.exp(returns.cumsum()) * start
    else:
        return (returns + 1).cumprod() * start


def to_returns(prices: TimeSeriesData, log: bool = False) -> TimeSeriesData:
    r"""
    Calculates the returns of a price series

    Parameters
    ----------
    prices
        price series
    log
        If True, calculate the log returns. :math:`\log{\frac{p1}{p0}}`
        Otherwise, calculates the arithmetic returns :math:`\frac{p1}{p0} - 1`.

    Returns
    -------
    TimeSeriesData
        Returns series
    """
    prices = to_time_series(prices)

    if log:
        return np.log(prices / prices.shift(1))
    else:
        return prices / prices.shift(1) - 1


def to_time_series(data, dates: Optional[DateTimes] = None, fail_policy='ignore') -> TimeSeriesData:
    """
    Casts the input data to a time series DataFrame or Series

    Parameters
    ----------
    data
        Input data. If it is 1 dimensional, will be cast into a Series object. If it
        2-dimensional, will be cast to a DataFrame. If data is already a DataFrame and
        has a column named 'date' (case insensitive), that column will
        become the index of the data. All the remaining data must be floats

    dates
        Sets the (datetime) index of the Series or DataFrame. If list of strings, each string
        must be convertible to a datetime object.

    fail_policy
        Policy to use when no dates can be set as index. Use one of [ignore, raise].
        If ignore, function will continue as is. If raise, an error will be raised
        when no dates can be set.

    Returns
    -------
    TimeSeriesData
        a time series Series or DataFrame
    """
    if not isinstance(data, abc.Iterable):
        raise ValueError('data must be a collection of data. Either a vector (Series) or a matrix (DataFrame)')

    dim = np.ndim(data)
    if dim == 1:
        data = pd.Series(data)
    elif dim == 2:
        data = pd.DataFrame(data)
    else:
        raise ValueError('data can only be 1 or 2 dimensional')

    fail_policy = fail_policy.lower()
    if fail_policy not in ('ignore', 'raise'):
        raise ValueError(f"Unknown <fail_policy>: {fail_policy}. Use one of [ignore, raise]")

    if dates is not None:
        dates = pd.to_datetime(dates).rename(None)
        data.index = dates
    elif dates is None and isinstance(data, pd.DataFrame):
        for col in data.columns:
            if str(col).lower() == 'date':
                data = data.copy()
                data.index = pd.to_datetime(data.pop(col)).rename(None)
                break
    elif fail_policy == 'raise':
        raise ValueError(f"unable to add date to data index")

    return data.astype(float)



import numpy as np
import pandas as pd

from ._file_handler import filepath

__all__ = ["load_cube", "load_etf", "load_hist", "load_smi"]


def load_cube(*, download=False) -> np.ndarray:
    """
    Loads a sample Monte Carlo simulation of 9 asset classes.

    The dimension of the cube is 80 * 1000 * 9. The first axis represents the time, the second
    represents the number of trials (simulations) and the third represents each asset class.

    Parameters
    ----------
    download: bool
        If True, forces the data to be downloaded again from the repository. Otherwise, loads the data from the
        stash folder

    Returns
    -------
    ndarray
        A data cube of simulated returns
    """
    return np.load(filepath('cube.npy', download))


def load_etf(*, date_as_index: bool = True, download=False) -> pd.DataFrame:
    """
    Dataset contains prices of 4 ETF ranging from 2001-06-15 to 2019-03-01.

    Parameters
    ----------
    date_as_index:
        If True, sets the first column as the index of the DataFrame

    download: bool
        If True, forces the data to be downloaded again from the repository. Otherwise, loads the data from the
        stash folder

    Returns
    -------
    DataFrame
        A data frame containing the prices of 4 ETF
    """
    fp = filepath('etf.csv', download)

    if date_as_index:
        df = pd.read_csv(fp, index_col=0, parse_dates=[0])
        df.index.name = df.index.name.strip()
    else:
        df = pd.read_csv(fp, parse_dates=[0])

    df.columns = df.columns.str.strip()
    for c in 'VBK', 'BND':
        df[c] = pd.to_numeric(df[c].str.strip())

    return df


def load_hist(*, date_as_index: bool = True, download=False) -> pd.DataFrame:
    """
    Dataset containing 20-years returns data from different asset classes spanning from 1988 to 2019.

    Parameters
    ----------
    date_as_index:
        If True, sets the first column as the index of the DataFrame

    download: bool
        If True, forces the data to be downloaded again from the repository. Otherwise, loads the data from the
        stash folder

    Returns
    -------
    DataFrame
        A data frame containing the prices of 4 ETF
    """
    fp = filepath('hist.csv', download)

    if date_as_index:
        df = pd.read_csv(fp, index_col=0, parse_dates=[0])
        df.index.name = df.index.name.strip()
    else:
        df = pd.read_csv(fp, parse_dates=[0])

    df.columns = df.columns.str.strip()
    return df


def load_smi(*, as_returns=False, download=False) -> pd.DataFrame:
    """
    Dataset contains the close prices of all 20 constituents of the Swiss Market Index (SMI) from
    2011-09-09 to 2012-03-28.

    Parameters
    ----------
    as_returns: bool
        If true, transforms the price data to returns data

    download: bool
        If True, forces the data to be downloaded again from the repository. Otherwise, loads the data from the
        stash folder

    Returns
    -------
    DataFrame
        A data frame of the closing prices of all 20 constituents of the Swiss Market Index
    """

    df = pd.read_csv(filepath('smi.csv', download), index_col=0, parse_dates=[0])
    if as_returns:
        df = df.pct_change().dropna()
    return df



import pandas as pd

from perfana.types import TimeSeriesData

__all__ = ['freq_to_scale', 'infer_frequency']


def freq_to_scale(freq: str):
    """Converts frequency to scale"""
    freq = str(freq).lower()
    if freq in ('d', 'day', 'daily'):
        return 252
    elif freq in ('w', 'week', 'weekly'):
        return 52
    elif freq in ('m', 'month', 'monthly'):
        return 12
    elif freq in ('s', 'semi-annual', 'semi-annually'):
        return 6
    elif freq in ('q', 'quarter', 'quarterly'):
        return 4
    elif freq in ('y', 'year', 'yearly', 'annual'):
        return 1
    else:
        raise ValueError(f"Unknown frequency: {freq}")


def infer_frequency(data: TimeSeriesData, fail_policy='raise'):
    """
    Infers the frequency (periodicity) of the time series

    Function works by taking the difference between the dates of each successive data point.
    It then identifies which is the most common difference, whether 2 dates differ by a day,
    a week or more and use that to determine the periodicity of the data.

    Parameters
    ----------
    data:
        time series pandas data object

    fail_policy:
        Action to do when frequency cannot be inferred. If set to 'raise', a TypeError will be raised.
        If set to 'ignore', a NoneType will be returned on failure.

    Returns
    -------
    str
        Frequency of data. If :code:`fail_policy` is ignore, returns None.

    Raises
    ------
    TypeError:
        If :code:`fail_policy` is set to 'raise' and frequency cannot be inferred
    """
    if not isinstance(data, (pd.DataFrame, pd.Series)):
        raise ValueError('<data> must be a pandas DataFrame or Series')

    fail_policy = fail_policy.lower()
    if fail_policy not in ('ignore', 'raise'):
        raise ValueError(f"Unknown <fail_policy>: {fail_policy}. Use one of [ignore, raise]")

    # custom inference
    n = len(data)
    time_delta = (data.index[1:] - data.index[:-1]).value_counts()

    delta = {
        'daily': sum(time_delta.get(f'{day} days', 0) for day in range(1, 6)),  # 1 to 5 days - daily data
        'weekly': time_delta.get('7 days', 0),
        'monthly': sum(time_delta.get(f'{day} days', 0) for day in (28, 29, 30, 31)),
        'quarterly': sum(time_delta.get(f'{day} days', 0) for day in range(89, 93)),
        'semi-annually': sum(time_delta.get(f'{day} days', 0) for day in range(180, 185)),  # 180-184 days - semi-annual
        'yearly': sum(time_delta.get(f'{day} days', 0) for day in range(360, 367))  # 360-366 - considered annual
    }

    if n < 30:  # less than 30 observations, just take key with max value
        return max(delta, key=lambda k: delta[k])

    for freq, count in delta.items():
        if count > 0.85 * n:
            # if over 85% of the data belongs to frequency category, select frequency
            return freq

    if fail_policy == 'raise':
        raise TypeError('could not infer periodicity of time series index')
    else:
        return None



class TimeIndexError(Exception):
    def __init__(self, *args):
        super().__init__("Could not determine time index for series. Please make sure that a date time index is "
                         "specified somewhere.", *args)


class TimeIndexMismatchError(Exception):
    def __init__(self, freq_a: str, freq_b: str, *args):
        err_msg = f"The time frequency for the data sets in comparison are not aligned. One is '{freq_a}' while the " \
            f"other is {freq_b}. Please make sure that they are equivalent for comparison. "

        super().__init__(err_msg, *args)



from typing import Optional, Union

import numpy as np
import pandas as pd

from perfana.conversions import to_time_series
from perfana.core.utils import freq_to_scale
from perfana.exceptions import TimeIndexError, TimeIndexMismatchError
from perfana.types import TimeSeriesData

__all__ = ['active_premium', 'annualized_returns', 'excess_returns', 'relative_returns']


def active_premium(ra: TimeSeriesData,
                   rb: TimeSeriesData,
                   freq: Optional[str] = None,
                   geometric=True,
                   prefixes=('AST', 'BMK')) -> TimeSeriesData:
    """
    The return on an investment's annualized return minus the benchmark's annualized return.

    Parameters
    ----------
    ra
        The assets returns vector or matrix

    rb
        The benchmark returns

    freq
        Frequency of the data. Use one of monthly, quarterly, semi-annually, yearly

    geometric
        If True, calculates the geometric returns. Otherwise, calculates the arithmetic returns

    prefixes
        Prefix to apply to overlapping column names in the left and right side, respectively. This is also applied
        when the column name is an integer (i.e. 0 -> AST_0). It is the default name of the Series data if there
        are no name to the Series

    Returns
    -------
    TimeSeriesData
        Active premium of each strategy against benchmark

    Examples
    --------
    >>> from perfana.datasets import load_etf
    >>> from perfana.core import active_premium
    # Get returns starting from the date where all etf has data
    >>> etf = load_etf().dropna().pa.to_returns().dropna()
    >>> active_premium(etf, etf)
              VBK       BND       VTI       VWO
    VBK  0.000000 -0.055385 -0.010407 -0.063939
    BND  0.055385  0.000000  0.044979 -0.008554
    VTI  0.010407 -0.044979  0.000000 -0.053532
    VWO  0.063939  0.008554  0.053532  0.000000
    >>> active_premium(etf.VBK, etf.BND)
              VBK
    BND  0.055385
    """
    ra = to_time_series(ra)
    rb = to_time_series(rb)

    if isinstance(ra, pd.Series):
        ra = pd.DataFrame(ra.rename(ra.name or prefixes[0]))

    if isinstance(rb, pd.Series):
        rb = pd.DataFrame(rb.rename(rb.name or prefixes[1]))

    freq = _determine_frequency(ra, rb, freq)

    res = pd.DataFrame()
    for ca, a in ra.iteritems():
        premium = {}
        if isinstance(ca, int):
            ca = f'{prefixes[0]}_{ca}'

        for cb, b in rb.iteritems():
            if isinstance(cb, int):
                cb = f'{prefixes[1]}_{cb}'

            premium[cb] = annualized_returns(a, freq, geometric) - annualized_returns(b, freq, geometric)
        res[ca] = pd.Series(premium)

    return res


def annualized_returns(r: TimeSeriesData,
                       freq: Optional[str] = None,
                       geometric=True) -> Union[float, pd.Series]:
    r"""
    Calculates the annualized returns from the data

    The formula for annualized geometric returns is formulated by raising the compound return to the number of
    periods in a year, and taking the root to the number of total observations:

    .. math::
        \prod_i^N(1 + r_i)^{\frac{s}{N}} - 1

    where :math:`s` is the number of observations in a year, and :math:`N` is the total number of observations.

    For simple returns (geometric=FALSE), the formula is:

    .. math::
        \frac{s}{N} \sum^N_i r_i


    Parameters
    ----------
    r
        Numeric returns series or data frame

    freq
        Frequency of the data. Use one of daily, weekly, monthly, quarterly, semi-annually, yearly

    geometric
        If True, calculates the geometric returns. Otherwise, calculates the arithmetic returns

    Returns
    -------
    float or Series
        Annualized returns

    Examples
    --------
    >>> from perfana.datasets import load_etf
    >>> from perfana.core import annualized_returns
    # Get returns starting from the date where all etf has data
    >>> etf = load_etf().dropna().pa.to_returns().dropna()
    VBK    0.091609
    BND    0.036224
    VTI    0.081203
    VWO    0.027670
    dtype: float64
    >>> annualized_returns(etf.VWO)
    0.02767037698144148
    """
    r = to_time_series(r).dropna()
    if freq is None:
        freq = r.pa.frequency

    scale = freq_to_scale(freq)

    if geometric:
        d = (r + 1).prod()
        return np.sign(d) * np.abs(d) ** (scale / len(r)) - 1
    else:  # arithmetic mean
        return r.mean() * scale


def excess_returns(ra: TimeSeriesData,
                   rb: TimeSeriesData,
                   freq: Optional[str] = None,
                   geometric=True) -> TimeSeriesData:
    r"""
    An average annualized excess return is convenient for comparing excess returns

    Excess returns is calculated by first annualizing the asset returns and benchmark returns stream. See the docs for
    `annualized_returns()` for more details. The geometric returns formula is:

    .. math::
        r_g = \frac{r_a - r_b}{1 + r_b}

    The arithmetic excess returns formula is:

    .. math::
        r_g = r_a - r_b

    Returns calculation will be truncated by the one with the shorter length. Also, annualized returns are calculated
    by the geometric annualized returns in both cases

    Parameters
    ----------
    ra
        The assets returns vector or matrix

    rb
        The benchmark returns. If this is a vector and the asset returns is a matrix, then all assets returns (columns)
        will be compared against this single benchmark. Otherwise, if this is a matrix, then assets will be compared
        to each individual benchmark (i.e. column for column)

    freq
        Frequency of the data. Use one of [daily, weekly, monthly, quarterly, semi-annually, yearly]

    geometric
        If True, calculates the geometric excess returns. Otherwise, calculates the arithmetic excess returns

    Returns
    -------
    TimeSeriesData
        Excess returns of each strategy against benchmark

    Examples
    --------
    >>> from perfana.datasets import load_etf
    >>> from perfana.core import excess_returns
    # Get returns starting from the date where all etf has data
    >>> etf = load_etf().dropna().pa.to_returns().dropna()
    >>> excess_returns(etf, etf.VBK)
    VBK    0.000000
    BND   -0.050737
    VTI   -0.009533
    VWO   -0.058573
    dtype: float64
    """
    ra = to_time_series(ra).dropna()
    rb = to_time_series(rb).dropna()

    n = min(len(ra), len(rb))
    ra, rb = ra.iloc[:n], rb.iloc[:n]

    if ra.ndim == rb.ndim and ra.shape != rb.shape:
        raise ValueError('The shapes of the asset and benchmark returns do not match!')

    freq = _determine_frequency(ra, rb, freq)

    ra = annualized_returns(ra, freq)
    rb = annualized_returns(rb, freq)

    return (ra - rb) / (1 + rb) if geometric else ra - rb


def relative_returns(ra: TimeSeriesData,
                     rb: TimeSeriesData,
                     prefixes=('AST', 'BMK')) -> TimeSeriesData:
    """
    Calculates the ratio of the cumulative performance for two assets through time

    Parameters
    ----------
    ra
        The assets returns vector or matrix

    rb
        The benchmark returns

    prefixes:
        Prefix to apply to overlapping column names in the left and right side, respectively. This is also applied
        when the column name is an integer (i.e. 0 -> AST_0). It is the default name of the Series data if there
        are no name to the Series

    Returns
    -------
    TimeSeriesData
        Returns a DataFrame of the cumulative returns ratio between 2 asset classes.
        Returns a Series if there is only 2 compared classes.

    Examples
    --------
    >>> from perfana.datasets import load_etf
    >>> from perfana.core import relative_returns
    # Get returns starting from the date where all etf has data
    >>> etf = load_etf().dropna().pa.to_returns().dropna()
    >>> relative_returns(etf.tail(), etf.VBK.tail())
                VBK/VBK   BND/VBK   VTI/VBK   VWO/VBK
    Date
    2019-02-25      1.0  0.996027  0.997856  1.009737
    2019-02-26      1.0  1.004013  1.002591  1.013318
    2019-02-27      1.0  0.997005  0.997934  1.000389
    2019-02-28      1.0  1.001492  1.001461  0.998348
    2019-03-01      1.0  0.987385  0.997042  0.988521
    """
    ra = to_time_series(ra)
    rb = to_time_series(rb)

    if isinstance(ra, pd.Series):
        ra = pd.DataFrame(ra.rename(ra.name or prefixes[0]))

    if isinstance(rb, pd.Series):
        rb = pd.DataFrame(rb.rename(rb.name or prefixes[1]))

    res = pd.DataFrame()
    for ca, a in ra.iteritems():
        for cb, b in rb.iteritems():
            df = (pd.merge(a, b, 'outer', left_index=True, right_index=True).dropna() + 1).cumprod()
            rel = df.iloc[:, 0] / df.iloc[:, 1]

            if isinstance(ca, int):
                ca = f'{prefixes[0]}_{ca}'
            if isinstance(cb, int):
                cb = f'{prefixes[1]}_{cb}'

            res = pd.merge(res, rel.rename(f'{ca}/{cb}'), 'outer', left_index=True, right_index=True)

    if res.shape[1] == 1:
        return res.iloc[:, 0]
    return res


def _determine_frequency(ra, rb, freq):
    if freq is None:
        fa, fb = ra.pa.frequency, rb.pa.frequency
        if fa is None and fb is None:
            raise TimeIndexError
        elif fa is None:
            freq = fb
        elif fb is None:
            freq = fa
        elif fa != fb:
            raise TimeIndexMismatchError(fa, fb)
        else:
            freq = fa

    return freq



from ._types import Frequency

__all__ = ['infer_frequency']


def infer_frequency(f: Frequency) -> int:
    assert isinstance(f, (str, int)), "Frequency must be either a string"

    if isinstance(f, str):
        f = f.lower()
        if f in ('week', 'weekly'):
            return 52
        if f in ('month', 'monthly'):
            return 12
        elif f in ('quarter', 'quarterly'):
            return 4
        elif f in ('semi-annual', 'semi-annually', 'semi-year', 'semi-yearly', 'half-year', 'half-yearly'):
            return 2
        elif f in ('annual', 'annually', 'year', 'yearly'):
            return 1

    assert f in (1, 2, 4, 12), \
        f"Unknown time frequency: {f}. For reference, 1 - year, 2 - semi-annual, 4 - quarter, 12 - month, 52: week"

    return f



from datetime import datetime
from typing import Iterable, Union

import numpy as np
import pandas as pd

__all__ = ['DateTimes', 'Scalar', 'TimeSeriesData', 'Vector', 'is_iterable']

DateTimes = Union[Iterable[str], Iterable[datetime]]
Scalar = Union[int, float]
Vector = Union[Iterable[Union[int, float]], np.ndarray, pd.Series]
TimeSeriesData = Union[pd.DataFrame, Vector]


def is_iterable(o):
    return hasattr(o, "__iter__")



from functools import lru_cache, wraps

import numpy as np

__all__ = ['lru_cache_ext']


def lru_cache_ext(f=None, maxsize=128, typed=False):
    """
    Extended lru cache that can cache mutable objects like numpy arrays. Does not work with dictionary arguments.

    Examples
    --------
    >>> from perfana.utils.cache import lru_cache_ext
    >>> import numpy as np
    >>> import time
    >>> @lru_cache_ext
    ... def calc(data: np.ndarray):
    ...     time.sleep(5)
    ...     return data.sum()
    >>> data = np.random.standard_exponential((10, 10, 10))
    >>> t = time.time()
    >>> calc(data)
    >>> print(round(time.time() - t, 5))
    5.01076
    >>> # repeat the same calculation
    >>> t = time.time()
    >>> calc(data)
    >>> print(round(time.time() - t, 5))
    0.009
    """

    def decorator(func):
        # Internal store that is used to pass data between the function called by the user and the calculator function
        # that the lru_cache works on.
        store = {}

        @wraps(func)
        def wrapper(*args, **kwargs):
            fn_hash = ""
            for a in args:
                fn_hash += _hash(a)

            for k, v in kwargs.items():
                fn_hash += _hash(v, k)
            store[fn_hash] = (args, kwargs)  # save all the arguments to the store
            return cached_caller(fn_hash)

        @lru_cache(maxsize, typed)
        def cached_caller(h):
            args, kwargs = store.pop(h)  # get the arguments data from store
            return func(*args, **kwargs)

        def _hash(data, k='') -> str:
            """This is used to calculate a pseudo-unique hash for the ndarray"""
            try:
                _h = str(hash(data))
            except TypeError:
                data = np.asarray(data)
                _h = str(hash(str(data)))

            return _h + k

        wrapper.cache_info = cached_caller.cache_info
        wrapper.cache_clear = cached_caller.cache_clear

        return wrapper

    if f is None:
        return decorator
    else:
        return decorator(f)



from setuptools import find_packages, setup

import versioneer

PACKAGE_NAME = 'perfana'

cmdclass = versioneer.get_cmdclass()

install_requires = [
    'copulae >=0.4',
    'matplotlib',
    'numpy',
    'pandas >=0.23',
    'scipy >=1.2'
]

setup(
    name=PACKAGE_NAME,
    license='MIT',
    version=versioneer.get_version().split('+')[0],
    description='Toolbox for performance and portfolio analytics',
    author='Daniel Bok',
    author_email='daniel.bok@outlook.com',
    packages=find_packages(include=['perfana', 'perfana.*']),
    url='https://github.com/DanielBok/perfana',
    cmdclass=cmdclass,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Financial and Insurance Industry',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'
    ],
    install_requires=install_requires,
    python_requires='>=3.6',
    zip_safe=False
)



from typing import NamedTuple, Union

import numpy as np

__all__ = ["Attribution", "Drawdown", "Frequency", "TailLoss"]

Attribution = NamedTuple("Attribution", [
    ("marginal", np.ndarray),
    ("percentage", np.ndarray),
])

Drawdown = NamedTuple("Drawdown", [
    ("average", float),
    ("paths", np.ndarray),
])

Frequency = Union[str, int]

TailLoss = NamedTuple("TailLoss", [
    ("prob", float),
    ("expected_loss", float)
])



from typing import Optional

import numpy as np
import pandas as pd

from perfana.conversions import to_time_series
from perfana.types import TimeSeriesData, Vector

__all__ = ["drawdown", "drawdown_summary"]


def drawdown(data: TimeSeriesData, weights: Vector = None, geometric=True, rebalance=True) -> pd.Series:
    """
    Calculates the drawdown at each time instance.

    If data is DataFrame-like, weights must be specified. If data is Series-like, weights
    can be left empty.

    Parameters
    ----------
    data
        The assets returns vector or matrix

    weights
        Weights of the portfolio. This must be 1 dimensional and must match the dimension of the data's
        last axis.

    geometric
        If True, calculates the geometric mean, otherwise, calculates the arithmetic mean.

    rebalance
        If True, portfolio is assumed to be rebalanced at every step.

    Returns
    -------
    Series
        Drawdown at each time instance

    Examples
    --------
    >>> from perfana.datasets import load_hist
    >>> from perfana.core import drawdown
    >>> hist = load_hist().iloc[:, :7]
    >>> weights = [0.25, 0.18, 0.24, 0.05, 0.04, 0.13, 0.11]
    >>> drawdown(hist, weights).min()
    -0.4007984968456346
    >>> drawdown(hist.iloc[:, 0]).min()
    -0.5491340502573534

    .. plot:: plots/core_drawdown.py
        :include-source:
    """
    data = to_time_series(data)
    weights = np.ravel(weights)

    assert data.ndim in (1, 2), "returns can only be a series or a dataframe"
    if data.ndim == 2:
        assert weights is not None, "weight must be specified if returns is a dataframe"
        weights = np.ravel(weights)

        if rebalance:
            cum_ret = (data @ weights + 1).cumprod() if geometric else (data @ weights).cumsum() + 1
        else:
            cum_ret = (data + 1).cumprod() @ weights if geometric else data.cumsum() @ weights + 1
    else:
        cum_ret = (data + 1).cumprod() if geometric else data.cumsum() + 1

    dd = cum_ret / cum_ret.expanding().max() - 1
    setattr(dd, "is_drawdown", True)
    return dd


def drawdown_summary(data: TimeSeriesData, weights: Vector = None, geometric=True, rebalance=True, *,
                     top: Optional[int] = 5) -> pd.DataFrame:
    """
    A summary of each drawdown instance. Output is ranked by depth of the drawdown.

    If data is DataFrame-like, weights must be specified. If data is Series-like, weights
    can be left empty.

    Parameters
    ----------
    data
        The assets returns vector or matrix

    weights
        Weights of the portfolio. This must be 1 dimensional and must match the dimension of the data's
        last axis.

    geometric
        If True, calculates the geometric mean, otherwise, calculates the arithmetic mean.

    rebalance
        If True, portfolio is assumed to be rebalanced at every step.

    top
        If None, returns all episodes. If specified, returns the top `n` episodes ranked by the depth
        of drawdown.

    Returns
    -------
    DataFrame
        A data frame summarizing each drawdown episode

    Examples
    --------
    >>> from perfana.datasets import load_hist
    >>> from perfana.core import drawdown_summary
    >>> hist = load_hist().iloc[:, :7]
    >>> weights = [0.25, 0.18, 0.24, 0.05, 0.04, 0.13, 0.11]
    >>> drawdown_summary(hist, weights)
           Start     Trough        End  Drawdown  Length  ToTrough  Recovery
    0 2007-11-30 2009-02-28 2014-02-28 -0.400798      76        16        60
    1 2000-04-30 2003-03-31 2004-02-29 -0.203652      47        36        11
    2 1990-01-31 1990-11-30 1991-05-31 -0.150328      17        11         6
    3 1998-04-30 1998-10-31 1999-06-30 -0.149830      15         7         8
    4 1994-02-28 1995-03-31 1996-01-31 -0.132766      24        14        10
    >>> drawdown_summary(hist.iloc[:, 0])
           Start     Trough        End  Drawdown  Length  ToTrough  Recovery
    0 2007-11-30 2009-02-28 2014-05-31 -0.549134      79        16        63
    1 2000-04-30 2003-03-31 2006-12-31 -0.474198      81        36        45
    2 1990-01-31 1990-09-30 1994-01-31 -0.286489      49         9        40
    3 1998-08-31 1998-09-30 1999-01-31 -0.148913       6         2         4
    4 2018-10-31 2018-12-31 2019-03-31 -0.130014       6         3         3
    """
    if not hasattr(data, 'is_drawdown') or not data.is_drawdown:
        data = drawdown(data, weights, geometric, rebalance)

    info = {
        "Drawdown": [],
        "Start": [],
        "Trough": [],
        "End": [],
    }

    # use indices to store date diff initially as it isn't easy to determine the unit of time
    # (days, months, etc). Also, heuristics like round(days / 30) breaks easily when date lengths
    # are long (month day diff and leap years break)
    prior_sign = 1 if data[0] >= 0 else 0
    min_, from_, current_ = 0, 0, data[0]

    # index, drawdown
    for i, dd in enumerate(data):
        sign = 0 if dd < 0 else 1

        if sign == prior_sign:
            if dd < current_:
                current_ = dd
                min_ = i
        else:
            info["Drawdown"].append(current_)
            info["Start"].append(from_)
            info["Trough"].append(min_)
            info["End"].append(i)
            min_, from_, current_ = i, i, dd
            prior_sign = sign

    info["Drawdown"].append(current_)
    info["Start"].append(from_)
    info["Trough"].append(min_)
    info["End"].append(len(data) - 1)

    # calculate various drawdown lengths
    df = pd.DataFrame(info)
    df["Length"] = df["End"] - df["Start"] + 1
    df["ToTrough"] = df["Trough"] - df["Start"] + 1
    df["Recovery"] = df["End"] - df["Trough"]

    # sort by drawdown
    df = df.sort_values('Drawdown').reset_index(drop=True)

    # convert indices to date
    df["Start"] = data.index[df["Start"]]
    df["Trough"] = data.index[df["Trough"]]
    df["End"] = data.index[df["End"]]

    # reorder columns
    df = df[df.Drawdown < 0][['Start', 'Trough', 'End', 'Drawdown', 'Length', 'ToTrough', 'Recovery']]

    assert top is None or (isinstance(top, int) and top > 0), "If top is not None, it must be a positive integer"

    if top is None:
        return df

    return df.head(top)
